/* vi: set sw=2 ts=8: */
/*****************************************************************************
*!
*! FILE NAME  : fwinfo.c
*!
*! DESCRIPTION: Functions for reading from fwinfo area in flash
*!              or fimage file.
*!
*! ---------------------------------------------------------------------------
*! (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
*!
*! This program is free software; you can redistribute it and/or modify
*! it under the terms of the GNU Lesser General Public License as published by
*! the Free Software Foundation; either version 2.1 of the License, or
*! (at your option) any later version.
*!
*! This program is distributed in the hope that it will be useful,
*! but WITHOUT ANY WARRANTY; without even the implied warranty of
*! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*! GNU Lesser General Public License for more details.
*!
*! You should have received a copy of the GNU Lesser General Public License
*! along with this program; if not, write to the Free Software
*! Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*!
*!***************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 8)
#include <mtd/mtd-user.h>
#else
#include <linux/mtd/mtd.h>
#endif

#include "fwinfo.h"

#define FWINFO_PARTITION "/dev/part/rescue"

/* 
 * For NOR flash images where the image info area is not present,
 * use this default info instead.
 * May be set on command line (by the makefile, ultimately from 
 * AXIS_CONFIG variables), or failing that, is set here to old, well-known 
 * constants.
 */

#ifndef FWINFO_NOR_DEFAULT_PTABLE_ADDR
#define FWINFO_NOR_DEFAULT_PTABLE_ADDR 0x10000
#endif

#ifndef FWINFO_NOR_DEFAULT_SECTORSIZE
#define FWINFO_NOR_DEFAULT_SECTORSIZE 0x10000 /* for NOR flashes we use */
#endif

#ifndef FWINFO_NOR_DEFAULT_TOTALSIZE
#define FWINFO_NOR_DEFAULT_TOTALSIZE 8
#endif

static const struct fwinfo nor_default_info = 
{
  .magic = FWINFO_MAGIC_INITIALIZED,
  .major = FWINFO_MAJOR, /* always latest version */
  .minor = FWINFO_MINOR, /* always latest version */
  .bootblkpar_flash_addr = 0xc000, /* previously hardcoded in bootblocktool */
  .bootblkpar_flash_end = 0x10000, /* previously hardcoded in bootblocktool */
  .ptable_flash_addr = FWINFO_NOR_DEFAULT_PTABLE_ADDR,
  .ptable_img_addr = FWINFO_NOR_DEFAULT_PTABLE_ADDR,
  .media_type = FWINFO_MEDIA_TYPE_NORFLASH,
  .media_erasesize = FWINFO_NOR_DEFAULT_SECTORSIZE,
  .media_pagesize = 0,
  .media_totalsize_mb = FWINFO_NOR_DEFAULT_TOTALSIZE,
  .hwids = HWID /* note: not space padded */
};

int 
read_flash_fwinfo(const char *device, struct fwinfo *flash_fwinfo)
{
  int flashd;
  off_t seek_status;
  int status;
  struct mtd_info_user mtd_info;
  int media_norflash;

  if (flash_fwinfo == NULL)
    goto bad_noclose;

  if (device == NULL || device[0] == '\0')
    device = FWINFO_PARTITION;

  flashd = open(device, O_RDONLY);
  if (flashd < 0) {
    perror("Opening flash device for fwinfo read");
    goto bad_noclose;
  }

  seek_status = lseek(flashd, FWINFO_LOCATION, SEEK_SET);
  if (seek_status == (off_t) -1) {
    perror("Seeking to fwinfo location in flash");
    goto bad;
  }

  status = ioctl(flashd, MEMGETINFO, &mtd_info);
  if (status < 0) {
    perror("Reading flash type");
    goto bad;
  }

  media_norflash = (mtd_info.type == MTD_NORFLASH);

  status = read(flashd, flash_fwinfo, sizeof(*flash_fwinfo));
  if (status < 0)
    perror("Reading fwinfo from flash");
  if (status != sizeof(*flash_fwinfo))
    goto bad;

  close(flashd);

  if (flash_fwinfo->magic == FWINFO_MAGIC_UNINITIALIZED)
    goto bad_noclose;
  
  if (flash_fwinfo->magic != FWINFO_MAGIC_INITIALIZED) {
    if (media_norflash) /* use NOR defaults */
      memcpy(flash_fwinfo, &nor_default_info, sizeof(nor_default_info));
    else
      goto bad_noclose;
  }

  return 0;

  /* Error exit */
bad:
  close(flashd);
bad_noclose:
  return -1;
}


int 
read_fimage_fwinfo(const char *fimage, struct fwinfo *fimage_fwinfo)
{
  if (fimage == NULL || fimage_fwinfo == NULL)
    return -1;

  memcpy(fimage_fwinfo, &fimage[FWINFO_LOCATION], sizeof(*fimage_fwinfo));

  if (fimage_fwinfo->magic == FWINFO_MAGIC_UNINITIALIZED)
    return -1;

  /* Assume NOR fimage if we can't find fwinfo area */
  if (fimage_fwinfo->magic != FWINFO_MAGIC_INITIALIZED)
    memcpy(fimage_fwinfo, &nor_default_info, sizeof(nor_default_info));

  return 0;
}

/******************************* END OF FILE ********************************/
